@@ -255,17 +255,18 @@ static int fluid_sffile_read_wav(SFData *sf, unsigned int start, unsigned int en
  * @param filename Path to the file to check
  * @return TRUE if it could be a SF2, SF3 or DLS file, FALSE otherwise
  *
+ * This function uses regular <code>fopen()</code>, <code>fread()</code> and <code>fseek()</code> to identify known Soundfont formats.
  * If fluidsynth was built with DLS support, this function will also identify DLS files.
  *
- * @note This function only checks whether header(s) in the RIFF chunk are present.
- * A call to fluid_synth_sfload() might still fail.
+ * @note This function only checks whether certain RIFF chunks are present in the file.
+ * A call to fluid_synth_sfload() might still fail, as it imposes much stricter structural checks.
  */
 int fluid_is_soundfont(const char *filename)
 {
     FILE    *fp;
     uint32_t fcc;
     int      retcode = FALSE;
-    const char* err_msg;
+    const char *err_msg;
 
     do
     {
@@ -300,20 +301,24 @@ int fluid_is_soundfont(const char *filename)
         }
 
         retcode = (fcc == SFBK_FCC);
+
         if(retcode)
         {
             break;  // seems to be SF2, stop here
         }
+
 #ifdef LIBINSTPATCH_SUPPORT
         else
         {
             IpatchFileHandle *fhandle = ipatch_file_identify_open(filename, NULL);
+
             if(fhandle != NULL)
             {
                 retcode = (ipatch_file_identify(fhandle->file, NULL) == IPATCH_TYPE_DLS_FILE);
                 ipatch_file_close(fhandle);
             }
         }
+
 #endif
     }
     while(0);
@@ -337,7 +342,7 @@ SFData *fluid_sffile_open(const char *fname, const fluid_file_callbacks_t *fcbs)
 
     if(!(sf = FLUID_NEW(SFData)))
     {
-        FLUID_LOG(FLUID_ERR, "Out of memory");
+        FLUID_LOG(FLUID_PANIC, "Out of memory");
         return NULL;
     }
 
@@ -356,7 +361,7 @@ SFData *fluid_sffile_open(const char *fname, const fluid_file_callbacks_t *fcbs)
 
     if(sf->fname == NULL)
     {
-        FLUID_LOG(FLUID_ERR, "Out of memory");
+        FLUID_LOG(FLUID_PANIC, "Out of memory");
         goto error_exit;
     }
 
@@ -453,6 +458,7 @@ void fluid_sffile_close(SFData *sf)
     SFInst *inst;
 
     fluid_rec_mutex_destroy(sf->mtx);
+
     if(sf->sffd)
     {
         sf->fcbs->fclose(sf->sffd);
@@ -720,7 +726,7 @@ static int process_info(SFData *sf, int size)
                 if((chunk.id != ICMT_FCC && chunk.size > 256) || (chunk.size > 65536) || (chunk.size % 2))
                 {
                     FLUID_LOG(FLUID_ERR, "INFO sub chunk %.4s has invalid chunk size of %d bytes",
-                              (char*)&chunk.id, chunk.size);
+                              (char *)&chunk.id, chunk.size);
                     return FALSE;
                 }
             }
@@ -739,7 +745,7 @@ static int process_info(SFData *sf, int size)
             /* alloc for chunk fcc and da chunk */
             if(!(item.fcc = FLUID_MALLOC(chunk.size + sizeof(uint32_t) + 1)))
             {
-                FLUID_LOG(FLUID_ERR, "Out of memory");
+                FLUID_LOG(FLUID_PANIC, "Out of memory");
                 return FALSE;
             }
 
@@ -861,19 +867,19 @@ static int pdtahelper(SFData *sf, unsigned int expid, unsigned int reclen, SFChu
 
     if(chunk->id != expid)
     {
-        FLUID_LOG(FLUID_ERR, "Expected PDTA sub-chunk '%.4s' found invalid id instead", (char*)&expid);
+        FLUID_LOG(FLUID_ERR, "Expected PDTA sub-chunk '%.4s' found invalid id instead", (char *)&expid);
         return FALSE;
     }
 
     if(chunk->size % reclen)  /* valid chunk size? */
     {
-        FLUID_LOG(FLUID_ERR, "'%.4s' chunk size is not a multiple of %d bytes", (char*)&expid, reclen);
+        FLUID_LOG(FLUID_ERR, "'%.4s' chunk size is not a multiple of %d bytes", (char *)&expid, reclen);
         return FALSE;
     }
 
     if((*size -= chunk->size) < 0)
     {
-        FLUID_LOG(FLUID_ERR, "'%.4s' chunk size exceeds remaining PDTA chunk size", (char*)&expid);
+        FLUID_LOG(FLUID_ERR, "'%.4s' chunk size exceeds remaining PDTA chunk size", (char *)&expid);
         return FALSE;
     }
 
@@ -1006,7 +1012,7 @@ static int load_phdr(SFData *sf, unsigned int size)
         /* load all preset headers */
         if((preset = FLUID_NEW(SFPreset)) == NULL)
         {
-            FLUID_LOG(FLUID_ERR, "Out of memory");
+            FLUID_LOG(FLUID_PANIC, "Out of memory");
             return FALSE;
         }
 
@@ -1099,7 +1105,7 @@ static int load_pbag(SFData *sf, int size)
 
             if((z = FLUID_NEW(SFZone)) == NULL)
             {
-                FLUID_LOG(FLUID_ERR, "Out of memory");
+                FLUID_LOG(FLUID_PANIC, "Out of memory");
                 return FALSE;
             }
 
@@ -1234,7 +1240,7 @@ static int load_pmod(SFData *sf, int size)
 
                 if((m = FLUID_NEW(SFMod)) == NULL)
                 {
-                    FLUID_LOG(FLUID_ERR, "Out of memory");
+                    FLUID_LOG(FLUID_PANIC, "Out of memory");
                     return FALSE;
                 }
 
@@ -1298,7 +1304,7 @@ int load_pgen(SFData *sf, int size)
     SFPreset *preset;
     SFGenAmount genval;
     unsigned short genid;
-    int level, skip, drop, discarded;
+    int level, skip, drop, discarded, z;
 
     preset_list = sf->preset;
 
@@ -1309,6 +1315,7 @@ int load_pgen(SFData *sf, int size)
         /* traverse through all presets */
         discarded = FALSE;
         zone_list = preset->zone;
+        z = 0;
 
         /* traverse preset's zones */
         while(zone_list)
@@ -1326,7 +1333,7 @@ int load_pgen(SFData *sf, int size)
 
                 if((size -= SF_GEN_SIZE) < 0)
                 {
-                    FLUID_LOG(FLUID_ERR, "Preset generator chunk size mismatch");
+                    FLUID_LOG(FLUID_ERR, "Preset generator chunk size mismatch, rejecting SoundFont as structurally defective.");
                     return FALSE;
                 }
 
@@ -1344,6 +1351,9 @@ int load_pgen(SFData *sf, int size)
                     else
                     {
                         skip = TRUE;
+                        FLUID_LOG(FLUID_DBG,
+                                  "Discarding out of order generator KeyRange in preset '%s' of zone %d (must be the first generator per SoundFont spec 8.1.2).",
+                                  preset->name, z);
                     }
                 }
                 else if(genid == GEN_VELRANGE)
@@ -1358,6 +1368,9 @@ int load_pgen(SFData *sf, int size)
                     else
                     {
                         skip = TRUE;
+                        FLUID_LOG(FLUID_DBG,
+                                  "Discarding out of order generator VelRange in preset '%s' of zone %d (must only be preceded by the KeyRange generator per SoundFont spec 8.1.2).",
+                                  preset->name, z);
                     }
                 }
                 else if(genid == GEN_INSTRUMENT)
@@ -1375,10 +1388,21 @@ int load_pgen(SFData *sf, int size)
                         /* generator valid? */
                         READW(sf, genval.sword);
                         dup = find_gen_by_id(genid, zone->gen);
+
+                        if(dup)
+                        {
+                            FLUID_LOG(FLUID_DBG,
+                                      "Duplicate generator %s in preset '%s' of zone %d found. Only the last occurrence is kept, previous one(s) will be dropped.",
+                                      fluid_gen_name(genid), preset->name, z);
+                        }
                     }
                     else
                     {
                         skip = TRUE;
+                        FLUID_LOG(FLUID_DBG,
+                                  "Generator %s encountered in preset '%s' of zone %d is either not valid for a preset zone or unknown; discarding.",
+                                  fluid_gen_name(genid), preset->name, z);
+
                     }
                 }
 
@@ -1389,7 +1413,7 @@ int load_pgen(SFData *sf, int size)
                         /* if gen ! dup alloc new */
                         if((g = FLUID_NEW(SFGen)) == NULL)
                         {
-                            FLUID_LOG(FLUID_ERR, "Out of memory");
+                            FLUID_LOG(FLUID_PANIC, "Out of memory");
                             return FALSE;
                         }
 
@@ -1422,7 +1446,7 @@ int load_pgen(SFData *sf, int size)
                 }
 
                 /* GEN_INSTRUMENT should be the last generator */
-                if (level == 3)
+                if(level == 3)
                 {
                     break;
                 }
@@ -1437,11 +1461,12 @@ int load_pgen(SFData *sf, int size)
                 /* advance to next zone before deleting the current list element */
                 zone_list = fluid_list_next(zone_list);
 
-                FLUID_LOG(FLUID_WARN, "Preset '%s': Discarding invalid global zone",
-                            preset->name);
+                FLUID_LOG(FLUID_WARN, "Preset '%s': Discarding invalid global zone (global zones must appear first in presets per SoundFont spec 7.3)",
+                          preset->name);
                 preset->zone = fluid_list_remove(preset->zone, zone);
                 delete_zone(zone);
 
+                z++;
                 /* we have already advanced the zone_list pointer, so continue with next zone */
                 continue;
             }
@@ -1454,21 +1479,35 @@ int load_pgen(SFData *sf, int size)
 
                 if((size -= SF_GEN_SIZE) < 0)
                 {
-                    FLUID_LOG(FLUID_ERR, "Preset generator chunk size mismatch");
+                    FLUID_LOG(FLUID_ERR, "Preset generator chunk size mismatch, rejecting SoundFont as structurally defective.");
                     return FALSE;
                 }
 
+                if(gen_list->data)
+                {
+                    FLUID_LOG(FLUID_DBG,
+                              "Generator %s in preset '%s' of zone %d appears after SampleID, in violation of SoundFont spec 8.1.2; discarding.",
+                              fluid_gen_name(((SFGen *)(gen_list->data))->id), preset->name, z);
+                }
+                else
+                {
+                    FLUID_LOG(FLUID_DBG,
+                              "Empty generator in preset '%s' of zone %d; discarding.",
+                              preset->name, z);
+                }
+
                 FSKIP(sf, SF_GEN_SIZE);
                 SLADVREM(zone->gen, gen_list);
             }
 
+            z++;
             zone_list = fluid_list_next(zone_list);
         }
 
         if(discarded)
         {
             FLUID_LOG(FLUID_WARN,
-                      "Preset '%s': Some invalid generators were discarded",
+                      "Preset '%s': Some invalid generators were discarded, potentially resulting in a different timbre. Run fluidsynth in verbose mode for detailed information.",
                       preset->name);
         }
 
@@ -1523,7 +1562,7 @@ static int load_ihdr(SFData *sf, unsigned int size)
         /* load all instrument headers */
         if((inst = FLUID_NEW(SFInst)) == NULL)
         {
-            FLUID_LOG(FLUID_ERR, "Out of memory");
+            FLUID_LOG(FLUID_PANIC, "Out of memory");
             return FALSE;
         }
 
@@ -1610,7 +1649,7 @@ static int load_ibag(SFData *sf, int size)
 
             if((z = FLUID_NEW(SFZone)) == NULL)
             {
-                FLUID_LOG(FLUID_ERR, "Out of memory");
+                FLUID_LOG(FLUID_PANIC, "Out of memory");
                 return FALSE;
             }
 
@@ -1746,7 +1785,7 @@ static int load_imod(SFData *sf, int size)
 
                 if((m = FLUID_NEW(SFMod)) == NULL)
                 {
-                    FLUID_LOG(FLUID_ERR, "Out of memory");
+                    FLUID_LOG(FLUID_PANIC, "Out of memory");
                     return FALSE;
                 }
 
@@ -1799,7 +1838,7 @@ int load_igen(SFData *sf, int size)
     SFInst *inst;
     SFGenAmount genval;
     unsigned short genid;
-    int level, skip, drop, discarded;
+    int level, skip, drop, discarded, z;
 
     inst_list = sf->inst;
 
@@ -1810,6 +1849,7 @@ int load_igen(SFData *sf, int size)
 
         discarded = FALSE;
         zone_list = inst->zone;
+        z = 0;
 
         /* traverse this instrument's zones */
         while(zone_list)
@@ -1828,7 +1868,7 @@ int load_igen(SFData *sf, int size)
 
                 if((size -= SF_GEN_SIZE) < 0)
                 {
-                    FLUID_LOG(FLUID_ERR, "IGEN chunk size mismatch");
+                    FLUID_LOG(FLUID_ERR, "IGEN chunk size mismatch, rejecting SoundFont as structurally defective.");
                     return FALSE;
                 }
 
@@ -1846,6 +1886,9 @@ int load_igen(SFData *sf, int size)
                     else
                     {
                         skip = TRUE;
+                        FLUID_LOG(FLUID_DBG,
+                                  "Discarding out of order generator KeyRange in instrument '%s' of zone %d (must be the first generator per SoundFont spec 8.1.2).",
+                                  inst->name, z);
                     }
                 }
                 else if(genid == GEN_VELRANGE)
@@ -1860,6 +1903,9 @@ int load_igen(SFData *sf, int size)
                     else
                     {
                         skip = TRUE;
+                        FLUID_LOG(FLUID_DBG,
+                                  "Discarding out of order generator VelRange in instrument '%s' of zone %d (must only be preceded by the KeyRange generator per SoundFont spec 8.1.2).",
+                                  inst->name, z);
                     }
                 }
                 else if(genid == GEN_SAMPLEID)
@@ -1867,6 +1913,15 @@ int load_igen(SFData *sf, int size)
                     /* sample is last gen */
                     level = 3;
                     READW(sf, genval.uword);
+
+                    if(fluid_list_next(gen_list))
+                    {
+                        FLUID_LOG(FLUID_DBG,
+                                  "Generator SampleID in instrument '%s' of zone %d is followed by additional generators. "
+                                  "Per SoundFont spec 8.1.2, SampleID must be the last generator in the zone. "
+                                  "All subsequent generators will be discarded.",
+                                  inst->name, z);
+                    }
                 }
                 else
                 {
@@ -1877,10 +1932,20 @@ int load_igen(SFData *sf, int size)
                         /* gen valid? */
                         READW(sf, genval.sword);
                         dup = find_gen_by_id(genid, zone->gen);
+
+                        if(dup)
+                        {
+                            FLUID_LOG(FLUID_DBG,
+                                      "Duplicate generator %s in instrument '%s' of zone %d found. Only the last occurrence is kept, previous one(s) will be dropped.",
+                                      fluid_gen_name(genid), inst->name, z);
+                        }
                     }
                     else
                     {
                         skip = TRUE;
+                        FLUID_LOG(FLUID_DBG,
+                                  "Generator %s in instrument '%s' of zone %d is invalid for instruments per SoundFont spec 8.1.2. Discarding.",
+                                  fluid_gen_name(genid), inst->name, z);
                     }
                 }
 
@@ -1891,7 +1956,7 @@ int load_igen(SFData *sf, int size)
                         /* if gen ! dup alloc new */
                         if((g = FLUID_NEW(SFGen)) == NULL)
                         {
-                            FLUID_LOG(FLUID_ERR, "Out of memory");
+                            FLUID_LOG(FLUID_PANIC, "Out of memory");
                             return FALSE;
                         }
 
@@ -1914,6 +1979,16 @@ int load_igen(SFData *sf, int size)
                     FSKIPW(sf);
                 }
 
+                if(drop)
+                {
+                    if(dup)
+                    {
+                        FLUID_LOG(FLUID_DBG,
+                                  "Dropping previous instance of generator %s in instrument '%s' of zone %d due to duplicate definition (last one kept).",
+                                  fluid_gen_name(genid), inst->name, z);
+                    }
+                }
+
                 if(!drop)
                 {
                     gen_list = fluid_list_next(gen_list);    /* next gen */
@@ -1924,7 +1999,7 @@ int load_igen(SFData *sf, int size)
                 }
 
                 /* GEN_SAMPLEID should be last generator */
-                if (level == 3)
+                if(level == 3)
                 {
                     break;
                 }
@@ -1939,11 +2014,12 @@ int load_igen(SFData *sf, int size)
                 /* advance to next zone before deleting the current list element */
                 zone_list = fluid_list_next(zone_list);
 
-                FLUID_LOG(FLUID_WARN, "Instrument '%s': Discarding invalid global zone",
-                            inst->name);
+                FLUID_LOG(FLUID_WARN, "Instrument '%s': Discarding invalid global zone (global zones must appear first in instruments per SoundFont spec 7.7).",
+                          inst->name);
                 inst->zone = fluid_list_remove(inst->zone, zone);
                 delete_zone(zone);
 
+                z++;
                 /* we have already advanced the zone_list pointer, so continue with next zone */
                 continue;
             }
@@ -1956,21 +2032,35 @@ int load_igen(SFData *sf, int size)
 
                 if((size -= SF_GEN_SIZE) < 0)
                 {
-                    FLUID_LOG(FLUID_ERR, "Instrument generator chunk size mismatch");
+                    FLUID_LOG(FLUID_ERR, "Instrument generator chunk size mismatch, rejecting SoundFont as structurally defective.");
                     return FALSE;
                 }
 
+                if(gen_list->data)
+                {
+                    FLUID_LOG(FLUID_DBG,
+                              "Generator %s in instrument '%s' of zone %d appears after SampleID, in violation of SoundFont spec 8.1.2; discarding.",
+                              fluid_gen_name(((SFGen *)(gen_list->data))->id), inst->name, z);
+                }
+                else
+                {
+                    FLUID_LOG(FLUID_DBG,
+                              "Empty generator in instrument '%s' of zone %d; discarding.",
+                              inst->name, z);
+                }
+
                 FSKIP(sf, SF_GEN_SIZE);
                 SLADVREM(zone->gen, gen_list);
             }
 
+            z++;
             zone_list = fluid_list_next(zone_list); /* next zone */
         }
 
         if(discarded)
         {
             FLUID_LOG(FLUID_WARN,
-                      "Instrument '%s': Some invalid generators were discarded",
+                      "Instrument '%s': Some invalid generators were discarded, audible glitches are to be expected! Run fluidsynth in verbose mode for detailed information.",
                       inst->name);
         }
 
@@ -2023,9 +2113,10 @@ static int load_shdr(SFData *sf, unsigned int size)
     {
         if((p = FLUID_NEW(SFSample)) == NULL)
         {
-            FLUID_LOG(FLUID_ERR, "Out of memory");
+            FLUID_LOG(FLUID_PANIC, "Out of memory");
             return FALSE;
         }
+
         p->idx = i;
 
         sf->sample = fluid_list_prepend(sf->sample, p);
@@ -2183,7 +2274,7 @@ static int valid_inst_genid(unsigned short genid)
 
     for(i = 0; i < FLUID_N_ELEMENTS(invalid_inst_gen); i++)
     {
-        if (invalid_inst_gen[i] == genid)
+        if(invalid_inst_gen[i] == genid)
         {
             return FALSE;
         }
@@ -2204,7 +2295,7 @@ static int valid_preset_genid(unsigned short genid)
 
     for(i = 0; i < FLUID_N_ELEMENTS(invalid_preset_gen); i++)
     {
-        if (invalid_preset_gen[i] == genid)
+        if(invalid_preset_gen[i] == genid)
         {
             return FALSE;
         }
@@ -2220,7 +2311,7 @@ static int fluid_sffile_read_wav(SFData *sf, unsigned int start, unsigned int en
     char *loaded_data24 = NULL;
     unsigned int num_samples;
 
-    fluid_return_val_if_fail((end + 1) > start , -1);
+    fluid_return_val_if_fail((end + 1) > start, -1);
 
     num_samples = (end + 1) - start;
 
@@ -2240,23 +2331,17 @@ static int fluid_sffile_read_wav(SFData *sf, unsigned int start, unsigned int en
     }
 
     loaded_data = FLUID_ARRAY(short, num_samples);
+
     if(loaded_data == NULL)
     {
-        FLUID_LOG(FLUID_ERR, "Out of memory");
+        FLUID_LOG(FLUID_PANIC, "Out of memory");
         goto error_exit_unlock;
     }
 
-    FLUID_LOG(FLUID_DBG, "ftell(): %llu, fread(): %ld bytes", sf->fcbs->ftell(sf->sffd), num_samples*sizeof(short));
+    FLUID_LOG(FLUID_DBG, "ftell(): %llu, fread(): %ld bytes", sf->fcbs->ftell(sf->sffd), num_samples * sizeof(short));
+
     if(sf->fcbs->fread(loaded_data, num_samples * sizeof(short), sf->sffd) == FLUID_FAILED)
     {
-#if FLUID_VERSION_CHECK(FLUIDSYNTH_VERSION_MAJOR, FLUIDSYNTH_VERSION_MINOR, FLUIDSYNTH_VERSION_MICRO) < FLUID_VERSION_CHECK(2,2,0)
-        if((int)(num_samples * sizeof(short)) < 0)
-        {
-            FLUID_LOG(FLUID_INFO,
-                      "This SoundFont seems to be bigger than 2GB, which is not supported in this version of fluidsynth. "
-                      "You need to use at least fluidsynth 2.2.0");
-        }
-#endif
         FLUID_LOG(FLUID_ERR, "Failed to read sample data");
         goto error_exit_unlock;
     }
@@ -2288,9 +2373,10 @@ static int fluid_sffile_read_wav(SFData *sf, unsigned int start, unsigned int en
         }
 
         loaded_data24 = FLUID_ARRAY(char, num_samples);
+
         if(loaded_data24 == NULL)
         {
-            FLUID_LOG(FLUID_ERR, "Out of memory reading 24-bit sample data");
+            FLUID_LOG(FLUID_PANIC, "Out of memory reading 24-bit sample data");
             goto error24_exit;
         }
 
@@ -2380,11 +2466,13 @@ static sf_count_t sfvio_seek(sf_count_t offset, int whence, void *user_data)
 
     new_offset += data->start;
     fluid_rec_mutex_lock(sf->mtx);
-    if (data->start <= new_offset && new_offset <= data->end &&
-        sf->fcbs->fseek(sf->sffd, new_offset, SEEK_SET) != FLUID_FAILED)
+
+    if(data->start <= new_offset && new_offset <= data->end &&
+            sf->fcbs->fseek(sf->sffd, new_offset, SEEK_SET) != FLUID_FAILED)
     {
         data->offset = new_offset - data->start;
     }
+
     fluid_rec_mutex_unlock(sf->mtx);
 
 fail:
@@ -2410,19 +2498,21 @@ static sf_count_t sfvio_read(void *ptr, sf_count_t count, void *user_data)
     }
 
     fluid_rec_mutex_lock(sf->mtx);
-    if (sf->fcbs->fseek(sf->sffd, data->start + data->offset, SEEK_SET) == FLUID_FAILED)
+
+    if(sf->fcbs->fseek(sf->sffd, data->start + data->offset, SEEK_SET) == FLUID_FAILED)
     {
         FLUID_LOG(FLUID_ERR, "This should never happen: fseek failed in sfvoid_read()");
         count = 0;
     }
     else
     {
-        if (sf->fcbs->fread(ptr, count, sf->sffd) == FLUID_FAILED)
+        if(sf->fcbs->fread(ptr, count, sf->sffd) == FLUID_FAILED)
         {
             FLUID_LOG(FLUID_ERR, "Failed to read compressed sample data");
             count = 0;
         }
     }
+
     fluid_rec_mutex_unlock(sf->mtx);
 
     data->offset += count;
@@ -2476,7 +2566,8 @@ static int fluid_sffile_read_vorbis(SFData *sf, unsigned int start_byte, unsigne
 
     /* Seek to sfdata.start, the beginning of Ogg Vorbis data in Soundfont */
     sfvio_seek(0, SEEK_SET, &sfdata);
-    if (sfdata.offset != 0)
+
+    if(sfdata.offset != 0)
     {
         FLUID_LOG(FLUID_ERR, "Failed to seek to compressed sample position");
         return -1;
@@ -2518,7 +2609,7 @@ static int fluid_sffile_read_vorbis(SFData *sf, unsigned int start_byte, unsigne
 
     if(!wav_data)
     {
-        FLUID_LOG(FLUID_ERR, "Out of memory");
+        FLUID_LOG(FLUID_PANIC, "Out of memory");
         goto error_exit;
     }
 
